@@ -1,27 +1,5 @@
 from enum import Enum
 
-try:
-    from dataclasses import dataclass
-
-    @dataclass
-    class TokenData:
-        access_token: str
-        expires_in: int
-        expires_at: int
-        refresh_expires_in: int
-        refresh_token: str
-        token_type: str
-        not_before_policy: int
-        session_state: str
-        scope: str
-
-
-except ImportError:
-
-    class TokenData(dict):
-        def __getattr__(self, item):
-            return self.get(item)
-
 
 class NodeType(Enum):
     OBJECT = "OBJECT"
