@@ -36,17 +36,42 @@ def stringify(v: Union[None, int, str, Dict, List]):
     )
 
 
+query_function_include = (
+    "declare function local:q($i, $q) {"
+    "let $m := for $k in jn:keys($q) return if (not(empty($i=>$k))) then deep-equal($i=>$k, $q=>$k) else false()"
+    "return empty(index-of($m, false()))"
+    "};"
+)
+
+upsert_function_include = (
+    "declare %updating function local:upsert-fields($r, $u) {"
+    "for $key in bit:fields($u) return if (empty($r=>$key)) then insert json $u into $r"
+    " else replace json value of $r=>$key with $u=>$key"
+    "};"
+)
+
+update_function_include = (
+    "declare %updating function local:update-fields($r, $u) {"
+    "for $key in bit:fields($u) return replace json value of $r=>$key with $u=>$key"
+    "};"
+)
+
+
 class JsonStoreBase(ABC):
+    __slots__ = ("db_name", "db_type", "name", "_client", "_auth")
+
     def __init__(
         self,
         db_name: str,
         name: str,
         client: Union[SyncClient, AsyncClient],
         auth: Auth,
+        root: str = "",
     ):
         self.db_name = db_name
         self.db_type = DBType.JSON
         self.name = name
+        self.root = root
         self._client = client
         self._auth = auth
 
@@ -58,7 +83,7 @@ class JsonStoreBase(ABC):
         :return: an emtpy ``str`` or an empty ``Awaitable[str]``.
         """
         insert_dict = dumps(insert_dict)
-        query = f"append json jn:parse('{insert_dict}') into jn:doc('{self.db_name}','{self.name}')"
+        query = f"append json jn:parse('{insert_dict}') into jn:doc('{self.db_name}','{self.name}'){self.root}"
         return self._client.post_query({"query": query})
 
     def insert_many(
@@ -72,7 +97,7 @@ class JsonStoreBase(ABC):
         """
         insert_list = dumps(insert_list)
         query = (
-            f"let $doc := jn:doc('{self.db_name}','{self.name}')"
+            f"let $doc := jn:doc('{self.db_name}','{self.name}'){self.root}"
             f"for $i in jn:parse('{insert_list}') return append json $i into $doc"
         )
         return self._client.post_query({"query": query})
@@ -85,14 +110,18 @@ class JsonStoreBase(ABC):
         """
         return self._client.resource_exists(self.db_name, self.db_type, self.name)
 
-    def create(self) -> Union[str, Awaitable[str]]:
+    def create(self, data: str = "[]") -> Union[str, Awaitable[str]]:
         """
         Creates the store, will overwrite the store if it already exists.
 
+        :param data: data with which to initialize the store
         :return: will return the string "[]". If in async mode, an awaitable that resolves this string.
         """
         return self._client.create_resource(
-            self.db_name, self.db_type, self.name, "[]",
+            self.db_name,
+            self.db_type,
+            self.name,
+            data,
         )
 
     def resource_history(self) -> Union[List[Commit], Awaitable[List[Commit]]]:
@@ -144,15 +173,6 @@ class JsonStoreBase(ABC):
             parse_revision(revision, params)
         return self._client.read_resource(self.db_name, self.db_type, self.name, params)
 
-    @staticmethod
-    def _prepare_query_dict(query_dict: Dict) -> str:
-        query_list = ["where"]
-        for k, v in query_dict.items():
-            query_list.append(f"deep-equal($i=>{stringify(k)}, {stringify(v)}) and")
-        if len(query_list) == 1:
-            return ""
-        return " ".join(query_list)[:-4]
-
     def _prepare_find_all(
         self,
         query_dict: Dict,
@@ -164,25 +184,33 @@ class JsonStoreBase(ABC):
         end_result_index: Optional[int] = None,
     ):
         if revision is None:
-            query_list = [f"for $i in jn:doc('{self.db_name}','{self.name}')"]
+            query_list = [
+                query_function_include,
+                f"for $i in jn:doc('{self.db_name}','{self.name}'){self.root}",
+            ]
         elif isinstance(revision, datetime):
             query_list = [
+                query_function_include,
                 "for $i in bit:array-values(jn:open"
-                f"('{self.db_name}','{self.name}',xs:dateTime('{revision.isoformat()}')))"
+                f"('{self.db_name}','{self.name}',xs:dateTime('{revision.isoformat()}')){self.root})",
             ]
         else:
             query_list = [
-                f"for $i in bit:array-values(jn:doc('{self.db_name}','{self.name}',{revision}))"
-            ]
-        query_list.append(self._prepare_query_dict(query_dict))
-        return_obj = "".join(
-            [
-                "return {$i",
-                ",'nodeKey': sdb:nodekey($i)" if node_key else "",
-                ",'hash': sdb:hash($i)}" if hash else "}",
+                query_function_include,
+                f"for $i in bit:array-values(jn:doc('{self.db_name}','{self.name}',{revision})){self.root}",
             ]
+        query_list.append(f"where local:q($i, {stringify(query_dict)})")
+        return_obj = (
+            "return {$i,'nodeKey': sdb:nodekey($i),'hash': sdb:hash($i)}"
+            if node_key and hash
+            else "return {$i,'nodeKey': sdb:nodekey($i)}"
+            if node_key
+            else "return {$i,'hash': sdb:hash($i)}"
+            if hash
+            else "return {$i}"
         )
-        query_string = " ".join([*query_list, return_obj])
+        query_list.append(return_obj)
+        query_string = " ".join(query_list)
         if projection is not None:
             if node_key:
                 projection.append("nodeKey")
@@ -230,18 +258,22 @@ class JsonStoreBase(ABC):
         raise NotImplementedError()
 
     def update_by_key(
-        self, node_key: int, update_dict: Dict[str, Union[List, Dict, str, int, None]],
+        self,
+        node_key: int,
+        update_dict: Dict[str, Union[List, Dict, str, int, None]],
+        upsert: bool = True,
     ) -> Union[str, Awaitable[str]]:
         """
 
         :param node_key: the nodeKey of the record to update
         :param update_dict: a dict of keys and matching values to replace in the given record
+        :param upsert: whether to insert if the field does not already exist
         :return:
         """
         query = (
+            f"{upsert_function_include if upsert else update_function_include}"
             f"let $rec := sdb:select-item(jn:doc('{self.db_name}','{self.name}'),{node_key}) "
-            f" let $update := {stringify(update_dict)}"
-            " for $key in bit:fields($update) return replace json value of $rec=>$key with $update=>$key"
+            f"return local:{'upsert' if upsert else 'update'}-fields($rec, {stringify(update_dict)})"
         )
         return self._client.post_query({"query": query})
 
@@ -249,17 +281,19 @@ class JsonStoreBase(ABC):
         self,
         query_dict: Dict,
         update_dict: Dict[str, Union[List, Dict, str, int, None]],
+        upsert: bool = True,
     ) -> Union[str, Awaitable[str]]:
         """
 
         :param query_dict: a ``dict`` of field names and their values to match against
         :param update_dict: a dict of keys and matching values to replace in the selected record
+        :param upsert: whether to insert if the field does not already exist
         :return:
         """
         query = (
-            f"for $i in jn:doc('{self.db_name}','{self.name}') {self._prepare_query_dict(query_dict)}"
-            f" let $update := {stringify(update_dict)}"
-            " for $key in bit:fields($update) return replace json value of $i=>$key with $update=>$key"
+            f"{query_function_include}{upsert_function_include if upsert else update_function_include}"
+            f"for $i in jn:doc('{self.db_name}','{self.name}'){self.root} where local:q($i, {stringify(query_dict)})"
+            f" return local:{'upsert' if upsert else 'update'}-fields($i, {stringify(update_dict)})"
         )
         return self._client.post_query({"query": query})
 
@@ -268,7 +302,7 @@ class JsonStoreBase(ABC):
     ) -> Union[str, Awaitable[str]]:
         """
 
-        :param node_key: the nodeKey of the record to update
+        :param node_key: the nodeKey of the record to delete
         :param fields: the keys of the fields of the record to delete
         :return:
         """
@@ -289,8 +323,9 @@ class JsonStoreBase(ABC):
         :return:
         """
         query = (
-            f"let $records := for $i in jn:doc('{self.db_name}','{self.name}')"
-            f" {self._prepare_query_dict(query_dict)} return $i"
+            f"{query_function_include}"
+            f"let $records := for $i in jn:doc('{self.db_name}','{self.name}'){self.root}"
+            f" where local:q($i, {stringify(query_dict)}) return $i"
             f" let $fields := {stringify(fields)}"
             f" for $i in $fields return delete json $records=>$i"
         )
@@ -303,14 +338,17 @@ class JsonStoreBase(ABC):
         :return:
         """
         query = (
-            f"let $doc := jn:doc('{self.db_name}','{self.name}')"
-            f" let $m := for $i at $pos in $doc {self._prepare_query_dict(query_dict)} return $pos - 1"
+            f"{query_function_include}"
+            f"let $doc := jn:doc('{self.db_name}','{self.name}'){self.root}"
+            f" let $m := for $i at $pos in $doc where local:q($i, {stringify(query_dict)}) return $pos - 1"
             " for $i in $m order by $i descending return delete json $doc[[$i]]"
         )
         return self._client.post_query({"query": query})
 
     def find_by_key(
-        self, node_key: Union[int, None], revision: Union[Revision, None] = None,
+        self,
+        node_key: Union[int, None],
+        revision: Union[Revision, None] = None,
     ):
         """
 
@@ -387,9 +425,7 @@ class JsonStoreSync(JsonStoreBase):
 
     def history(
         self, node_key: int, subtree: bool = True, revision: Optional[Revision] = None
-    ) -> Union[
-        List[SubtreeRevision], List[RevisionType],
-    ]:
+    ) -> Union[List[SubtreeRevision], List[RevisionType],]:
         return super().history(node_key, subtree, revision)["rest"]
 
     def resource_history(self) -> List[Commit]:
@@ -401,7 +437,9 @@ class JsonStoreSync(JsonStoreBase):
         return super().history_embed(node_key, revision)["rest"]
 
     def find_by_key(
-        self, node_key: Union[int, None], revision: Union[Revision, None] = None,
+        self,
+        node_key: Union[int, None],
+        revision: Union[Revision, None] = None,
     ):
         return super().find_by_key(node_key, revision)
 
@@ -454,6 +492,8 @@ class JsonStoreAsync(JsonStoreBase):
         return result["rest"]
 
     async def find_by_key(
-        self, node_key: Union[int, None], revision: Union[Revision, None] = None,
+        self,
+        node_key: Union[int, None],
+        revision: Union[Revision, None] = None,
     ):
         return await super().find_by_key(node_key, revision)
